/**
 * Project: PS5
 * Authors: @Logan Tucker and @Tucker Simpson
 * Training for viterbi algorithm, cs10, winter 2022
 *
 */

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.HashMap;
import java.util.Locale;

public class Train {

    // create storage
    HashMap<String, HashMap<String, Double>> EmissionMap;
    HashMap<String, HashMap<String, Double>> TransmissionMap;

    HashMap<String, Integer> eRowTotals;
    HashMap<String, Integer> tRowTotals;

    public Train(String textPathName, String posPathName) {
        getMaps(textPathName, posPathName);
    }

    /**
     * Takes a text file and a file of its corresponding parts of speech and creates Emission and Transmission maps
     * of probability
     * @param textPathName text file
     * @param posPathName parts of speech file
     */
    public void getMaps(String textPathName, String posPathName) {
        EmissionMap = new HashMap<>();
        TransmissionMap = new HashMap<>();

        eRowTotals = new HashMap<>();
        tRowTotals = new HashMap<>();

        BufferedReader textInput;
        BufferedReader posInput;

        // if files open without error
        try {

            textInput = new BufferedReader(new FileReader(textPathName));
            posInput = new BufferedReader(new FileReader(posPathName));

            // read the first line from each file
            String textLine = textInput.readLine();
            String posLine = posInput.readLine();

            // while there are unread lines in both files
            while (textLine != null && posLine != null) {

                // split each line by " " and hold words in a list of strings
                String[] words = textLine.split(" ");
                String[] pos = posLine.split(" ");

                for (int i = 0; i < words.length && i < pos.length; i++) {
                    String currWord = words[i].toLowerCase(Locale.ROOT);

                    // if this is the first time we have encountered a POS
                    if (!EmissionMap.containsKey(pos[i])) {

                        // create a key for this POS in both maps
                        EmissionMap.put(pos[i], new HashMap<String, Double>());
                        TransmissionMap.put(pos[i], new HashMap<String, Double>());

                        // create a place to store the row total for this POS (occurrences)
                        eRowTotals.put(pos[i], 0);
                        tRowTotals.put(pos[i], 0);
                    }

                    // if the current word has occurred as this POS before
                    if (EmissionMap.get(pos[i]).containsKey(currWord)) {

                        // increment its occurrences
                        EmissionMap.get(pos[i]).put(currWord, EmissionMap.get(pos[i]).get(currWord) + 1);
                    }

                    else {

                        // place the current word into the map with an occurrence of 1
                        EmissionMap.get(pos[i]).put(currWord, 1.0);
                    }

                    // increment row total for current POS (occurrences of this POS)
                    eRowTotals.put(pos[i], eRowTotals.get(pos[i]) + 1);

                    // if we are looking at the first word in the sentence account for special predecessor "#"
                    if (i == 0) {

                        // place "#" key into transition map if it does not already exist
                        if (!TransmissionMap.containsKey("#")) {
                            TransmissionMap.put("#", new HashMap<String, Double>());
                            tRowTotals.put("#", 0);
                        }

                        // if we have never seen the current transition before
                        if (!TransmissionMap.get("#").containsKey(pos[i])) {

                            // place the current transition into the map with an occurrence of 1
                            TransmissionMap.get("#").put(pos[i], 1.0);
                        }
                        else {

                            // increment the occurrences of the current transition (transitions from this POS)
                            TransmissionMap.get("#").put(pos[i], TransmissionMap.get("#").get(pos[i]) + 1);
                        }
                        // increment row total for current POS
                        tRowTotals.put("#", tRowTotals.get("#") + 1);
                    }
                    else {

                        // if the current transition has occurred before
                        if (TransmissionMap.get(pos[i-1]).containsKey(pos[i])) {

                            // increment its occurrences
                            TransmissionMap.get(pos[i-1]).put(pos[i], TransmissionMap.get(pos[i-1]).get(pos[i]) + 1);
                        }
                        else {
                            // place the current transition into the map with an occurrence of 1
                            TransmissionMap.get(pos[i-1]).put(pos[i], 1.0);
                        }

                        // increment row total for current POS (transitions from this POS)
                        tRowTotals.put(pos[i-1], tRowTotals.get(pos[i-1]) + 1);
                    }
                }

                // read next lines
                textLine = textInput.readLine();
                posLine = posInput.readLine();
            }

            // for both maps turn each occurrence value into a log probability
            for (String oK : EmissionMap.keySet()) {
                for (String iK : EmissionMap.get(oK).keySet()) {
                    EmissionMap.get(oK).put(iK, Math.log(EmissionMap.get(oK).get(iK) / eRowTotals.get(oK)));
                }
            }

            for (String oK : TransmissionMap.keySet()) {
                for (String iK : TransmissionMap.get(oK).keySet()) {
                    TransmissionMap.get(oK).put(iK, Math.log(TransmissionMap.get(oK).get(iK) / tRowTotals.get(oK)));
                }
            }
        }
        catch(Exception FileNotFoundException) {
            System.out.println("File not found");
        }
    }
}
